/*
 *	lymelib2db - conversion from library text files to a database for
 *	LyME on Palm devices.
 *	Based on lymelib2db.m, a similar utility for SysQuake and Matlab
 *
 *	Usage: lymelib2db [-n dbname] dbfile lib1 lib2 ...
 *
 *	Copyright 2001-2003, Calerga Sarl
 *	http://www.calerga.com
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#if defined(WIN32)
#	include <time.h>
#else
#	include <sys/time.h>
#	include <unistd.h>
#endif

static char const defaultDBName[] = "UntitledLyMELibDB";
static char const type[] = "Lml ";
static char const creator[] = "LyME";

static long palmTime(long y, long m, long d, long h, long min, long s)
{
	static long const mc[] = {0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334};
	return ((((y - 1904) * 365
			+ mc[m - 1]
			+ (y - 1901 + (m >= 3)) / 4
			+ d - 1) * 24 + h) * 60 + min) * 60 + s;
}

static void writePDBHeader(FILE *fp,
		char const *name,
		char const type[4], char const creator[4],
		int lengths[], int nLengths)
{
	char buf[32];
	int i;
	long pt, offset;
	
	/* name (32 characters, padded with zeros) */
	if (strlen(name) >= 32)
		fwrite(name, 1, 32, fp);
	else
	{
		fwrite(name, 1, strlen(name), fp);
		memset(buf, 0, 32);
		fwrite(buf, 1, 32 - strlen(name), fp);
	}
	
	/* attributes (8 = backup) and version (0) */
	fwrite("\010\000\000\000", 1, 4, fp);
	
	/* creation and modification date */
	{
		time_t t;
		struct tm *date;
		
		time(&t);
		date = localtime(&t);
		pt = palmTime(date->tm_year, date->tm_mon, date->tm_mday,
				date->tm_hour, date->tm_min, date->tm_sec);
	}
	
	buf[0] = pt >> 24;
	buf[1] = pt >> 16;
	buf[2] = pt >> 8;
	buf[3] = pt;
	
	fwrite(buf, 1, 4, fp);
	fwrite(buf, 1, 4, fp);
	
	/* last backup (none) */
	fwrite(buf + 4, 1, 4, fp);
	
	/* modification number */
	buf[8] = 1;
	fwrite(buf + 5, 1, 4, fp);
	
	/* appinfoId (none) */
	fwrite(buf + 4, 1, 4, fp);
	
	/* sortInfoId (none) */
	fwrite(buf + 4, 1, 4, fp);
	
	/* type */
	fwrite(type, 1, 4, fp);
	
	/* creator */
	fwrite(creator, 1, 4, fp);
	
	/* uniqueIDSeed (time) */
	fwrite(buf, 1, 4, fp);
	
	/* nextRecordListID (0) */
	fwrite(buf + 4, 1, 4, fp);
	
	/* length */
	buf[0] = nLengths >> 8;
	buf[1] = nLengths;
	fwrite(buf, 1, 2, fp);
	
	offset = 78 + 8 * nLengths;
	for (i = 0; i < nLengths; i++)
	{
		buf[0] = offset >> 24;	/* localChunkID */
		buf[1] = offset >> 16;
		buf[2] = offset >> 8;
		buf[3] = offset;
		buf[4] = 64;	/* attributes = dirty */
		buf[5] = i + 1 >> 16;	/* uniqueID */
		buf[6] = i + 1 >> 8;
		buf[7] = i + 1;
		fwrite(buf, 1, 8, fp);
		offset += lengths[i];
	}
}

int convEol2LF(FILE *fpIn, FILE *fpOut)
	/* convert end-of-line sequences in fpIn to LF, writing result to fpOut if !NULL
	   ret. length of result */
{
	int c, lenOut, lastWasCR;
	
	for (lenOut = lastWasCR = 0; ; )
	{
		c = fgetc(fpIn);
		if (c < 0)
			break;
		if (c == 10 && lastWasCR)
			lastWasCR = 0;
		else
		{
			if (c == 13)
			{
				c = 10;
				lastWasCR = 1;
			}
			else
				lastWasCR = 0;
			
			if (fpOut)
				fputc(c, fpOut);
			
			lenOut++;
		}
	}
	
	return lenOut;
}

int main(int argc, char **argv)
{
	char const *dbname = defaultDBName, *dbfile = NULL, **names = NULL;
	char **libs;
	int i, n, *nameLengths = NULL;
	int *lengths;
	FILE *fpIn, *fpOut = NULL;
	int status = 1;
	static const char usage[] = "Usage: lymelib2db [-n dbname] dbfile lib1 lib2 ...\n";
	
	/* decode arguments */
	for (i = 1; i < argc; i++)
		if (!strcmp(argv[i], "-n") && i + 1 < argc)
			dbname = argv[++i];
		else if (!dbfile)
			dbfile = argv[i];
		else if (argv[i][0] == '-')
		{
			fprintf(stderr, usage);
			exit(1);
		}
		else
			break;
	libs = argv + i;
	n = argc - i;
	if (n <= 0)
	{
		fprintf(stderr, usage);
		exit(1);
	}
	
	/* create array of lengths */
	if (i < argc)
	{
		lengths = (int *)malloc(n * sizeof(int));
		names = (char const **)malloc(n * sizeof(char const *));
		nameLengths = (int *)malloc(n * sizeof(int));
		if (!lengths || !names || !nameLengths)
		{
			fprintf(stderr, "Not enough memory.\n");
			exit(1);
		}
		
		for (i = 0; i < n; i++)
		{
			fpIn = fopen(libs[i], "rb");
			if (!fpIn)
			{
				fprintf(stderr, "Cannot open file \"%s\"\n.", libs[i]);
				goto cleanUp;
			}
			/* length of library file */
			lengths[i] = convEol2LF(fpIn, NULL);
			fclose(fpIn);
			/* make library name from filename */
			if (strchr(libs[i], '.'))
				nameLengths[i] = strrchr(libs[i], '.') - libs[i];
			else
				nameLengths[i] = strlen(libs[i]);
			for (names[i] = libs[i] + nameLengths[i];
					names[i] > libs[i]
						&& names[i][-1] != '/' && names[i][-1] != '\\';
					names[i]--)
				;
			nameLengths[i] -= names[i] - libs[i];
			/* add length of library name + 1 */
			lengths[i] += nameLengths[i] + 1;
		}
	}
	else
		lengths = NULL;
	
	/* create pdb file */
	fpOut = fopen(dbfile, "wb");
	if (!fpOut)
	{
		fprintf(stderr, "Cannot open file \"%s\".\n", dbfile);
		exit(1);
	}
	
	/* write header */
	writePDBHeader(fpOut, dbname, type, creator, lengths, n);
	
	/* write libraries */
	for (i = 0; i < n; i++)
	{
		fpIn = fopen(libs[i], "rb");
		if (!fpIn)
		{
			fprintf(stderr, "Cannot open file \"%s\"\n.", libs[i]);
			goto cleanUp;
		}
		fwrite(names[i], 1, nameLengths[i], fpOut);
		fwrite("\x0a", 1, 1, fpOut);
		(void)convEol2LF(fpIn, fpOut);
		fclose(fpIn);
	}
	
	status = 0;
	
	/* clean up */
cleanUp:
	if (fpOut)
		fclose(fpOut);
	if (lengths)
		free((void *)lengths);
	if (names)
		free((void *)names);
	if (nameLengths)
		free((void *)nameLengths);
	
	return status;
}
